"""The solution for the iris problem."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.datasets import load_iris
from sklearn.metrics import ConfusionMatrixDisplay, confusion_matrix
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.neighbors import KNeighborsClassifier


def compute_accuracy(y: np.ndarray, y_pred: np.ndarray) -> float:
    """Compute the accuracy of the prediction.

    The accuracy is the most common performance measure for classification tasks.
    Is is calculated as the ration of correct predictions to all predictions.

    Args:
        y (np.ndarray): The array with the ground truth labels.
        y_pred (np.ndarray): The array with the predicted labels.

    Returns:
        (float): The accuracy of the prediction.
    """
    accuracy = (y == y_pred).mean()
    return accuracy


def accuracy_most_frequent(y_train: np.ndarray, y_test: np.ndarray) -> float:
    """Compute the accuracy of the most frequent model.

    Most frequent models always predict the label that occurs the most in the train set.
    They belong to the class of so-called "dummy models", because the prediction is
    independent of the input. Such models usually serve as a baseline if
    no other models are available.

    Args:
        y_train (np.ndarray): The array with the training labels.
        y_test (np.ndarray): The array with the test labels.

    Returns:
        (float): The accuracy of the prediction by the most frequent model on the test labels.
    """
    # count occurrences for each class in train set
    classcount = np.bincount(y_train)
    # find most frequent label in train set
    most_frequent_class = classcount.argmax()
    # compute accuracy
    acc_most_frequent = (y_test == most_frequent_class).mean()
    return acc_most_frequent


# optional
def accuracy_stratified(y_train: np.ndarray, y_test: np.ndarray) -> float:
    """Compute the accuracy of the stratified model.

    A stratified model assigns random labels based on the ratio of the labels in the train set.

    Args:
        y_train (np.ndarray): The array with the training labels.
        y_test (np.ndarray): The array with the test labels.

    Returns:
        (float): The accuracy of the prediction by the stratified model on the test labels.
    """
    # count occurrences for each class in train set
    class_counts = np.bincount(y_train)
    # compute probabilities of each class in train set
    class_probs = class_counts / len(y_train)
    # predict with stratified model
    y_test_stratified_pred = np.random.choice(
        len(class_counts), size=y_test.shape[0], p=class_probs
    )
    # compute accuracy
    acc_stratified = (y_test == y_test_stratified_pred).mean()
    return acc_stratified


def cv_knearest_classifier(x_train: np.ndarray, y_train: np.ndarray) -> GridSearchCV:
    """Train and cross-validate a k-nearest neighbors classifier with the grid search.

    Define a k-nearest neighbors classifier. Use the grid search with the grid from 1 to 25
    and a 3-fold cross-validation to find the best value for the hyperparameter k.

    Args:
        x_train (np.ndarray): The training data.
        y_train (np.ndarray): The training labels.

    Returns:
        (GridSearchCV): The trained model that was cross-validated with the grid search.
    """
    # create new knn model
    knn2 = KNeighborsClassifier()
    # create dictionary of all values we want to test for n_neighbors
    param_grid = {"n_neighbors": np.arange(1, 25)}
    # use gridsearch and 3-fold cross-validation to test all values for n_neighbors
    knn_gscv = GridSearchCV(knn2, param_grid, cv=3)
    # fit model to data
    knn_gscv.fit(x_train, y_train)
    return knn_gscv


if __name__ == "__main__":
    # load iris dataset
    iris = load_iris()

    # print shape of data matrix and number of target entries
    print(iris.data.shape, iris.target.shape)
    # print names of labels and of features
    print(iris.target_names)  # names of labels
    print(iris.feature_names)  # names of features

    # (optional) use classes distribution (histogram) to check if iris dataset is balanced:
    # find out what the next two lines of code do
    temp = pd.Series(iris.target)
    target_str = temp.apply(lambda i: iris.target_names[i])
    # and use 'pandas.Series.hist' function to plot histogram
    ax = target_str.hist()
    plt.show()

    # (optional) use pandas 'scatter_matrix' to visualize some trends in data
    # represent iris as pandas data frame
    iris_df = pd.DataFrame(iris.data, columns=iris.feature_names)
    # create scatter matrix from dataframe, color by target; plot matrix
    ax = pd.plotting.scatter_matrix(
        iris_df,
        c=iris.target,
        marker="o",
        figsize=(8, 8),
        hist_kwds={"bins": 20},
        s=60,
        alpha=0.8,
    )
    plt.show()

    # create train and test split with the ratio 75:25 and print their dimensions
    xtrain, xtest, ytrain, ytest = train_test_split(
        iris.data, iris.target, train_size=0.75, random_state=29
    )  # 29
    print(xtrain.shape, ytrain.shape)  # dimension of train set
    print(xtest.shape, ytest.shape)  # dimension of test set

    # create a k-nn, with k = 1
    knn = KNeighborsClassifier(n_neighbors=1)

    # train the classifier on the train set
    knn.fit(xtrain, ytrain)

    # predict labels first for train and then for test data
    ytrain_pred = knn.predict(xtrain)  # predict on train set
    ytest_pred = knn.predict(xtest)  # predict on test set

    # use 'compute_accuracy' to evaluate your model by calculating accuracy on train and test set
    acc_train = compute_accuracy(ytrain, ytrain_pred)
    acc_test = compute_accuracy(ytest, ytest_pred)
    # print both accuracies
    print(f"Accuracy of the train set: {acc_train}")
    print(f"Accuracy of the test set: {acc_test}")

    # use 'accuracy_most_frequent' to compute accuracy of most frequent model
    acc_mf = accuracy_most_frequent(ytrain, ytest)
    # print result
    print(f"Accuracy of the most frequent model: {acc_mf}")

    # (optional) use 'accuracy_stratified' to compute and print accuracy of the stratified model
    np.random.seed(29)  # set random seed for reproducibility
    acc_strat = accuracy_stratified(ytrain, ytest)
    print(f"Stratified accuracy: {acc_strat}")

    # compute confusion matrix for test set
    cm_test = confusion_matrix(ytest, ytest_pred)

    # (optional) compute and print test set accuracy from confusion matrix
    acc_confusion = np.diag(cm_test).sum() / cm_test.sum()
    print(acc_confusion)

    # plot heatmap of confusion matrix for test set
    disp = ConfusionMatrixDisplay(cm_test, display_labels=iris.target_names)
    disp.plot()
    plt.show()

    # use 'cv_knearest_classifier' to perform cross validation with grid search
    knn_cv = cv_knearest_classifier(xtrain, ytrain)

    # print top performing n_neighbors value
    bestn = knn_cv.best_params_
    print(bestn)

    # print mean score for the top performing value of n_neighbors
    bestscore = knn_cv.best_score_
    print(bestscore)

    # get access to best estimator
    bestestimator = knn_cv.best_estimator_
    # calculate train and test accuracy on best estimator
    ytrain_pred2 = bestestimator.predict(xtrain)  # predict on train set
    ytest_pred2 = bestestimator.predict(xtest)  # predict on test set

    # evaluate your model by calculating accuracy on train and test set
    acc_train2 = compute_accuracy(ytrain, ytrain_pred2)
    acc_test2 = compute_accuracy(ytest, ytest_pred2)
    # print both accuracies
    print(f"Accuracy of the train set: {acc_train2}")
    print(f"Accuracy of the test set: {acc_test2}")

    # plot heatmap of the confusion matrix for test set
    cm_test2 = confusion_matrix(ytest, ytest_pred2)
    disp2 = ConfusionMatrixDisplay(cm_test2, display_labels=iris.target_names)
    disp2.plot()
    plt.show()

    # # not needed
    # # accuracy on train and test sets over parameter values
    # acc_train,acc_test = [],[]
    # for k in range(1,15):
    #     knn = KNeighborsClassifier(n_neighbors=k)
    #     knn.fit(xtrain, ytrain)
    #     acc_train.append((ytrain != knn.predict(xtrain)).mean())
    #     acc_test.append((ytest != knn.predict(xtest)).mean())
    #
    # # plot accuracies over n_neighbors
    # plt.plot(acc_train, label='Train misclassification')
    # plt.plot(acc_test, label='Test misclassification')
    # plt.legend()
    # plt.xlabel("n_neighbors")
    # plt.ylabel("misclassification")
    # plt.show()
