"""The nearest neighbour regression solution."""

from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.datasets import fetch_california_housing
from sklearn.metrics import mean_absolute_error, mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsRegressor
from sklearn.preprocessing import StandardScaler


def gs_knearest_regressor(
    xtrain: np.ndarray, ytrain: np.ndarray
) -> Tuple[KNeighborsRegressor, np.signedinteger, np.ndarray]:
    """Implement the grid search for k-nearest neighbors regressor.

    Split the train set into a train and a validation set (80:20) and train the new train set for k = 1 to 40.
    For each estimator calculate mean squared error on the val set and store the results in an array.
    Determine, which value of k should we take and train the regressor on the whole train set with
    the newly determined value of k.

    Args:
        xtrain (np.ndarray): The training data.
        ytrain (np.ndarray): The training labels.

    Returns:
        (Tuple[KNeighborsRegressor, signedinteger, np.ndarray]):
        KNeighborsRegressor: The model that was trained with the optimal number of k.
        signedinteger: k that produced the smallest mean squared error on the val set.
        np.ndarray: The array with means squared errors for k values between 1 and 39.
    """
    error_val = []
    for i in range(1, 40):
        xtrain2, xval, ytrain2, yval = train_test_split(
            xtrain, ytrain, test_size=0.20, random_state=29
        )
        knn = KNeighborsRegressor(n_neighbors=i)
        knn.fit(xtrain2, ytrain2)
        pred_i = knn.predict(xval)
        mse_val = mean_squared_error(yval, pred_i)
        error_val.append(mse_val)
    opt_num_neighbors = np.argmin(error_val) + 1
    knn = KNeighborsRegressor(n_neighbors=opt_num_neighbors)
    knn.fit(xtrain, ytrain)
    error_array = np.array(error_val)
    return knn, opt_num_neighbors, error_array


if __name__ == "__main__":
    # as_frame=True loads data in dataframe format, with other metadata besides it
    california_housing = fetch_california_housing(as_frame=True)

    # select only dataframe part
    df = california_housing.frame

    # assign MedHouseVal to y as labels and all other columns to x as features
    y = df["MedHouseVal"]
    x = df.drop(["MedHouseVal"], axis=1)

    # use 'describe' and 'print' to check if we have differences in feature measurements
    print(x.describe())

    # We see that the features have differences in the mean and the standard deviation. We are using an algorithm
    # based on a distance and distance-based algorithms suffer greatly from data that isn't
    # on the same scale, such as this data. The scale of the points will distort the real
    # distance between values. That means that we will need to scale the features on the train data.

    # split the data into test and train set, 80:20, use random_state = 29
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.20, random_state=29
    )

    # print dimensions of test and train data
    print(len(x))
    print(len(x_train))
    print(len(x_test))

    # use 'StandardScaler' on train data
    scaler = StandardScaler()
    scaler.fit(x_train)

    # scale both train and test data
    x_train = scaler.transform(x_train)
    x_test = scaler.transform(x_test)

    # we can now look at the scaled data
    col_names = [
        "MedInc",
        "HouseAge",
        "AveRooms",
        "AveBedrms",
        "Population",
        "AveOccup",
        "Latitude",
        "Longitude",
    ]
    scaled_df = pd.DataFrame(x_train, columns=col_names)
    print(scaled_df.describe().T)

    # define a k-nn regressor with k = 1
    regressor = KNeighborsRegressor(n_neighbors=1)

    # train the regressor and calculate mean squared error on the test set
    regressor.fit(x_train, y_train)

    y_pred = regressor.predict(x_test)

    mae = mean_absolute_error(y_test, y_pred)
    mse = mean_squared_error(y_test, y_pred)
    rmse = mean_squared_error(y_test, y_pred, squared=False)

    # print mean squared error
    print(f"mae: {mae}")
    print(f"mse: {mse}")
    print(f"rmse: {rmse}")

    print(regressor.score(x_test, y_test))

    y.describe()

    # Implement the function 'gs_knearest_regressor' to perform the grid search to
    # determine the best value for k for k-nearest neighbors regressor.
    # Use this function to get the best estimator, the best number for k and
    # the error array of the search. Print, which value of k should we take?
    knn_gs, num_neighbors, error_array = gs_knearest_regressor(x_train, y_train)
    print(f"neighbors: {num_neighbors}")
    # compute and print MSE of best estimator on test set
    pred_i_t = knn_gs.predict(x_test)
    mse_t = mean_squared_error(y_test, pred_i_t)
    print(f"mse test: {mse_t}")

    # plot mean squared error for k values between 1 and 39
    plt.figure(figsize=(12, 6))
    plt.plot(
        range(1, 40),
        error_array,
        color="red",
        linestyle="dashed",
        marker="o",
        markerfacecolor="blue",
        markersize=10,
        label="validation error",
    )
    plt.title("K Value MSE")
    plt.xlabel("K Value")
    plt.ylabel("Mean Squared Error")
    plt.legend()
    plt.show()
